from qgis.PyQt.QtCore import (QCoreApplication, QT_TRANSLATE_NOOP)
from qgis.core import (
  QgsCoordinateReferenceSystem,
  QgsProcessingAlgorithm,
  QgsProcessingParameterExtent,
  QgsProcessingParameterDistance,
  QgsProcessingParameterCrs, 
  QgsProcessingParameterFeatureSink,
  QgsProcessingParameterRasterDestination,
  QgsProcessingParameterString,
  QgsProcessingParameterBoolean,
  QgsProcessingParameterNumber,
  QgsReferencedRectangle,
  QgsRectangle
  )
from qgis import processing

from ..algutil.hriskutil import HrUtil
from ..algutil.hrisktile import HrQgsTile
from ..algutil.hriskpostprocessor import HrPostProcessor
from ..algutil.hriskvar import PostProcessors

import requests
import re
import os

class fetchdemrastersrtm(QgsProcessingAlgorithm):
  
  PARAMETERS = {  
    "FETCH_EXTENT": {
      "ui_func": QgsProcessingParameterExtent,
      "ui_args":{
        "description": QT_TRANSLATE_NOOP("fetchdemrastersrtm","Extent for fetching data")
      }
    },
    "TARGET_CRS": {
      "ui_func": QgsProcessingParameterCrs,
      "ui_args": {
        "description": QT_TRANSLATE_NOOP("fetchdemrastersrtm","Target CRS (Cartesian coordinates)")
      }
    },
    "BUFFER": {
      "ui_func": QgsProcessingParameterDistance,
      "ui_args": {
        "description": QT_TRANSLATE_NOOP("fetchdemrastersrtm","Buffer of the fetch area (using Target CRS)"),
        "defaultValue": 0.0,
        "parentParameterName": "TARGET_CRS"
      }
    },
    "VECTORIZE": {
      "ui_func": QgsProcessingParameterBoolean,
      "ui_args": {
        "description": QT_TRANSLATE_NOOP("fetchdemrasterja","Output elevation points as vector layer"),
        "defaultValue": False
      }
    },
    "LOAD_CREDENTIAL": {
      "ui_func": QgsProcessingParameterBoolean,
      "ui_args": {
        "description": QT_TRANSLATE_NOOP("fetchdemrastersrtm","Use saved username and password to login SRTM data system"),
        "defaultValue": False
      }
    },
    "USERNAME": {
      "ui_func": QgsProcessingParameterString,
      "ui_args": {
        "description": QT_TRANSLATE_NOOP("fetchdemrastersrtm","Username to login SRTM data system"),
        "optional": True
      }
    },
    "PASSWORD": {
      "ui_func": QgsProcessingParameterString,
      "ui_args": {
        "description": QT_TRANSLATE_NOOP("fetchdemrastersrtm","Password to login SRTM data system"),
        "optional": True
      }
    },
    "SAVE_CREDENTIAL": {
      "ui_func": QgsProcessingParameterBoolean,
      "ui_args": {
        "description": QT_TRANSLATE_NOOP("fetchdemrastersrtm","Save (overwrite) username and password to login SRTM data system"),
        "defaultValue": False
      }
    },
    "MAX_DOWNLOAD": {
      "ui_func": QgsProcessingParameterNumber,
      "ui_args": {
        "description": QT_TRANSLATE_NOOP("fetchdemrastersrtm","Maximum number of download"),
        "type": QgsProcessingParameterNumber.Integer,
        "defaultValue": 4
      }
    },
    "OUTPUT": {
      "ui_func": QgsProcessingParameterFeatureSink,
      "ui_args": {
        "description": QT_TRANSLATE_NOOP("fetchdemrastersrtm","Elevation points")
      }
    },
    "OUTPUT_RASTER": {
      "ui_func": QgsProcessingParameterRasterDestination,
      "ui_args": {
        "description": QT_TRANSLATE_NOOP("fetchdemrastersrtm","Elevation raster" )
      }
    }
  }  
  
  FETCH_BASE_URL = "https://e4ftl01.cr.usgs.gov/DP109/SRTM/SRTMGL1.003/2000.02.11/{lat_str}{long_str}.SRTMGL1.hgt.zip"
  FETCH_LOGIN_URL = "https://urs.earthdata.nasa.gov"
  FETCH_CRS = QgsCoordinateReferenceSystem("EPSG:4326")
  FETCH_GEOM_TYPE = "Point"
  FETCH_TILE_FAMILY = "long_lat"
  FETCH_TILE_PARAMETERS = {"xunit": 1.0, "yunit": 1.0, "xorig": 0.0, "yorig": 0.0}
    
  def __init__(self):
    super().__init__()
    self.UTIL = HrUtil(self)
  
  # initialize of the algorithm  
  def initAlgorithm(self, config):    
    (extent, target_crs) = self.UTIL.getExtentAndCrsUsingCanvas()
    self.UTIL.setDefaultValue("FETCH_EXTENT", extent)
    self.UTIL.setDefaultValue("TARGET_CRS", target_crs.authid())
    self.UTIL.initParameters()

  # execution of the algorithm
  def processAlgorithm(self, parameters, context, feedback): 
    
    
    self.UTIL.registerProcessingParameters(parameters, context, feedback)
    self.CURRENT_PROCESS = self.UTIL.parseCurrentProcess()
    # get target x-y CRS, to apply the buffer and determine the fetch area
    target_crs = self.parameterAsCrs(parameters, "TARGET_CRS", context)
    
    # check whether the target CRS is x-y coordinates
    self.UTIL.checkCrsAsCartesian(target_crs)
    
    # get the extent, using the target CRS
    fetch_extent = self.parameterAsExtent(
      parameters, "FETCH_EXTENT", context, 
      self.parameterAsCrs(parameters, "TARGET_CRS", context)
      )
    
    # get the buffer
    buffer = self.parameterAsDouble(parameters, "BUFFER",context)
    
    # get the fetch area, using the extent and buffer
    fetch_area = QgsReferencedRectangle(
      QgsRectangle(
        fetch_extent.xMinimum() - buffer,
        fetch_extent.yMinimum() - buffer,
        fetch_extent.xMaximum() + buffer,
        fetch_extent.yMaximum() + buffer
      ),
      target_crs
    )
    
    tile = HrQgsTile(
      crs = self.FETCH_CRS,
      **self.FETCH_TILE_PARAMETERS
    )
    
    long_lat_list = tile.cellXyIdx(fetch_area)
    
    
    fetch_args = {}
    for i, (long, lat) in enumerate(long_lat_list):
      long_str = f"{long:+04}".replace("+","E").replace("-","W")
      lat_str  = f"{lat:+03}".replace("+","N").replace("-","S")
      fetch_args[f"{i+1}/{len(long_lat_list)}"] = {
        "url": self.FETCH_BASE_URL.format(lat_str = lat_str, long_str = long_str)
      }
          
    if len(fetch_args) > self.parameterAsInt(parameters, "MAX_DOWNLOAD", context):
      feedback.reportError(self.tr("Too many downloads are required: ") + str(len(fetch_args)))
      raise Exception(self.tr("Too many downloads are required: ") + str(len(fetch_args)))
    
    # set login data and login
    srtm_session = requests.Session()
    srtm_token = re.search(
      r'<form[^>]*id="login".*?>[\s\S]*?<input[^>]*name="authenticity_token"[^>]*value="([^"]*)"[^>]*>', 
      srtm_session.get(self.FETCH_LOGIN_URL).text
    ).group(1)
    
    if self.parameterAsBool(parameters, "LOAD_CREDENTIAL", context):
      try:
        with open(os.path.join(os.path.dirname(__file__), "srtm_credential"), "r") as f:
          line1 = f.readline().strip()
          line2 = f.readline().strip()
          username = re.search(r"username: (.*)", line1).group(1)
          password = re.search(r"password: (.*)", line2).group(1)
      except:
        raise Exception(self.tr("Failed to load the credential"))
    else:
      username = self.parameterAsString(parameters, "USERNAME", context)
      password = self.parameterAsString(parameters, "PASSWORD", context)
    
    if self.parameterAsBool(parameters, "SAVE_CREDENTIAL", context):
      with open(os.path.join(os.path.dirname(__file__), "srtm_credential"), "w") as f:
        f.write("username: " + username + "\n")
        f.write("password: " + password + "\n")
    
    srtm_session.request(
      method = "POST",
      url = self.FETCH_LOGIN_URL + "/login",
      data = {
        "utf8": "✓",
        "authenticity_token": srtm_token,
        "username": username,
        "password": password
      }
    )
    
    fetch_results = self.UTIL.downloadFilesConcurrently(
      args=fetch_args, session=srtm_session,
      parallel=False
    )
    
    files_unpack = []
    for file in fetch_results.values():
      _, extracted_files = self.UTIL.extractArchive(file, extension=".zip", pattern = "\\.hgt$")
      files_unpack += extracted_files
    
    ras_raw = processing.run(
      "gdal:merge",{
        "INPUT": files_unpack,
        "OUTPUT": self.parameterAsOutputLayer(parameters, "OUTPUT_RASTER", context)
      },
      context = context,
      is_child_algorithm = True
    )["OUTPUT"]
    
    
    # # download files using the session info
    # self.fetchFeaturesFromWeb(parameters, context, feedback)
    
    if self.parameterAsBoolean(parameters, "VECTORIZE", context):
      # clip the raster because it is too large as a point vector
      try: # UNKNOWN error may be occured in the following process
        dem_raster_clipped = processing.run(
          "gdal:cliprasterbyextent", 
          {
            "INPUT": ras_raw,
            "PROJWIN": fetch_area,
            "OUTPUT": "TEMPORARY_OUTPUT"
          },
          context = context,
          is_child_algorithm = True
        )["OUTPUT"]
      
      except:
        dem_raster_clipped = ras_raw
      
    
      # create vector layer from the raster
      dem_raw = processing.run(
        "native:pixelstopoints",{
          "INPUT_RASTER": dem_raster_clipped,
          "RASTER_BAND": 1,
          "FIELD_NAME": "alti",
          "OUTPUT": "TEMPORARY_OUTPUT"
        },
        context = context,
        is_child_algorithm = True
      )["OUTPUT"]
      
      # CRS transform    
      dem_transformed = processing.run(
        "native:reprojectlayer", 
        {
          "INPUT": dem_raw,
          "TARGET_CRS": target_crs,
          "OUTPUT": "TEMPORARY_OUTPUT"
        },
        context = context,
        is_child_algorithm = True
      )["OUTPUT"]      
              
      # substitute self constant with the fetched vector layer
      dem_final = processing.run(
        "hrisk:initelevationpoint",{
          "INPUT": dem_transformed,
          "FIELD_USE_AS_HEIGHT": "alti",
          "TARGET_CRS": target_crs,
          "OVERWRITE": True,
          "OUTPUT": "TEMPORARY_OUTPUT"
        },
        context = context,
        is_child_algorithm = True
      )["OUTPUT"]
      
      dem_final_fts = context.getMapLayer(dem_final)
      new_fields = self.UTIL.newFieldsWithHistory(dem_final_fts.fields())
      (sink, dest_id) = self.parameterAsSink(
        parameters, "OUTPUT", context,
        new_fields, dem_final_fts.wkbType(), target_crs
      )
      self.UTIL.addFeaturesWithHistoryToSink(
        sink, dem_final_fts, new_fields,
        current_process=self.CURRENT_PROCESS
      )
      
      
      PostProcessors[dest_id] = HrPostProcessor(
        history=[self.CURRENT_PROCESS]
      )
      
    else:
      dest_id = None
    
    
    # register history and postprocessor for the raster output
    PostProcessors[ras_raw] = HrPostProcessor(
      history=[self.CURRENT_PROCESS], 
      color_args = {"coloring": "single_band_pseudo_color", "theme": "Greens", "opacity": 0.8},
      set_min_to_zero = True
    )  
    
    self.UTIL.registerPostProcessAlgorithm(context, PostProcessors)  
    
    return {"OUTPUT": dest_id, "OUTPUT_RASTER": ras_raw}   
    
  def name(self):
    return self.__class__.__name__
  
  def displayName(self):
    return self.tr("Elevation points (SRTM)")

  def group(self):
    return self.tr('Fetch geometries')

  def groupId(self):
    return 'fetchgeometry'

  def createInstance(self):
    return fetchdemrastersrtm()

  # placing here is necessary, when employing pylupdate
  def tr(self, string):
    return QCoreApplication.translate(self.__class__.__name__, string)
